unit U_GridSubdivisions_V3;
{Copyright  2016, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }
{Version 3: Display the solution and cleanup}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, StdCtrls;

type
  TGridDef = array of array of integer;
  TFigLocs = array of array of TPoint;
  TForm1 = class(TForm)
    Grid1: TStringGrid;
    SolveBtn: TButton;
    ResetBtn: TButton;
    Memo1: TMemo;
    FigNbrBox: TCheckBox;
    ProgressBox: TCheckBox;
    procedure FormActivate(Sender: TObject);
    procedure Grid1DrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure ResetBtnClick(Sender: TObject);
    procedure SolveBtnClick(Sender: TObject);
    Private
    Public
      grid:TGridDef;
      FigLocs:TFigLocs;
      procedure ReDrawGrid;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}
var
  {The figure ID numbers by row and column}
  DefVal: array[0..5,0..5] of integer =
         ((1,1,1,1,1,2),
          (1,1,1,2,2,2),
          (1,2,2,2,2,3),
          (1,4,5,5,5,3),
          (4,4,5,6,5,3),
          (4,4,6,6,3,3));


{*************** FormActivate *************}
procedure TForm1.FormActivate(Sender: TObject);
begin

  ReDrawGrid;
end;

{************ RedrawGrid ***************}
Procedure TForm1.RedrawGrid;
var  {fill the grid with numbers representing the areas}
  i,n, r,c:integer;
  nbrcells:array[0..5] of integer; {# of cells in each figure}
begin
  if figNbrbox.checked and (not progressbox.checked)
  then grid1.Font.size:=-16 else grid1.font.size:=-30;
  grid1.canvas.font.assign(grid1.font);
  setlength(grid,6,6);
  setlength(Figlocs,6, 36);
  for i:= 0 to 5 do nbrcells[i]:=0;
  for r:=0 to 5 do
  for c:=0 to 5 do
  begin
    n:=defval[r,c]-1;
    grid1.cells[c,r]:=inttostr(n); {fill the visible grid}
    grid[c,r]:=n;  {fill the internal representive of the grid}
    figlocs[n,nbrcells[n]]:=point(c,r);
    inc(nbrcells[n]);{count cells by figure nbr}
  end;
  for i:=0 to 5 do setlength(figlocs[i], nbrcells[i]);
  {stop here to check that grid and  figlocs are properly set up}
  //showmessage('Checkpoint 1');
end;

{************* Grid1DrawCell **************}
procedure TForm1.Grid1DrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  with grid1, canvas, rect do
  begin
    //pen.width:=1; pen.Color:=clgray
    fillrect(rect);
    pen.width:=3; pen.color:=clblue;
    If  (acol=0) or
    ((acol>0) and (cells[acol,arow][1]<>cells[acol-1,arow][1]))
    then
    begin moveto(left,top); lineto(left,bottom); end;

    If (arow=0) or
    ((arow>0) and (cells[acol,arow][1]<>cells[acol,arow-1][1]))
    then
    begin
      moveto(left,top); lineto(right,top);
    end;

    if acol=colcount-1 then
    begin
      moveto(right,top); lineto(right,bottom);
    end;
    if arow=rowcount-1 then
    begin
      moveto(left,bottom); lineto(right,bottom);
    end;
    //if progressbox.checked then
    begin
      if grid[acol,arow]=7
      then textout(left+10, top+2, '*')
      else if length(grid1.cells[acol,arow])>1
      then textout(left+10, top+2, '?')
      else if fignbrBox.Checked
      then textout(left+10, top+2, cells[acol,arow])
      else textout(left+10, top+2, ' ');
    end
    ;//else if fignbrbox.Checked
    //then textout(left+10, top+2, cells[acol,arow][1])

  end;
end;

procedure TForm1.ResetBtnClick(Sender: TObject);
begin
  RedrawGrid;
end;

{************* SolveBtnClick *************}
procedure TForm1.SolveBtnClick(Sender: TObject);
{Use a recursive search to place one star (*) in every section
 of the grid in such a way that there is one in each row and column and
 no two stars are adjacent in any direction.}

    {------------- ValidLoc ------------}
    function ValidLoc(n:integer; p:TPoint):boolean;
    var
      x,y:integer;
    begin {make sure that the other cells in this row
           and column and any adjoing cells do not already
           contain a star}
      result:=true;
      for x:=0 to 5 do {check columns}
      if (x<>p.x) and (grid[x,p.y]>5) then
      begin
        result:=false;
        exit;
      end;

      for y:=0 to 5 do {check rows}
      if (y<>p.y) and (grid[p.x,y]>5) then
      begin
        result:=false;
        exit;
      end;

      {check adjacent diagonals}
      if p.x>0 then
      begin {check column left (West) diagonals}
        if ((p.y>0)and (grid[p.X-1,p.Y-1]> 5)) {NorthWest}
        or ((p.Y < 5) and (grid[p.X-1,p.Y+1]> 5)) {SouthWest}
        then
        begin
          result:=false;
          exit;
        end;
      end;
      if (p.x<5)then {Check right column(East) diagonals}
      begin
        if ((p.y>0) and (grid[p.X+1,p.Y-1]> 5)) {NorthEast}
        or ((p.Y < 5)and(grid[p.X+1,p.Y+1]> 5)) {SouthEast}
        then
        begin
          result:=false;
          exit;
        end;
      end;
    end;

    {--------------- PlaceNextStar ----------------}
    function PlacenextStar(starNbr:integer; nextgrid:TGridDef):boolean;
    {Recursive function to place stars on the grid according to the rules given}
    var
      i,c,r:integer;
      savecell:string;
      p:TPoint;
    begin
      result:=false;
      if starnbr<=5 {no message for final call (starNbr=6)}
      then memo1.Lines.add('Trying figure '+inttostr(starnbr));
      If starnbr>5 then
      begin  {we have a solution!}
        grid1.invalidate; {draw it}
        showmessage('We have a solution!');
        result:=true;
        exit;
      end
      else
      begin {here we go - place a star in the next valid position of the given
             figure number and call this function again with the next figure number
             If we cannot find a valid place in this figure for a star, return False
             to the caller.  If we had placed a star and called this routine, the
             false return means that our star is not correctly placed, retract the
             placement and continue looking for the next valid position.
             }
        for i:= 0 to high(figlocs[starnbr]) do
        with grid1 do
        begin
          p:=figlocs[starnbr,i];
          memo1.Lines.Add(format('Cell at col %d, row %d',
                                   [p.x,p.y]));
          with grid1 do
          begin
            savecell:=cells[p.x,p.y];
            Cells[p.X,p.Y]:=Cells[p.x,p.y]+'?';
            grid1.update;
            If progressbox.checked then sleep(1000);
          end;
          if validloc(starnbr,p) then
          begin
            memo1.Lines.add('is valid');
            nextgrid[p.x, p.y]:=7;
            savecell:=cells[p.x,p.y];
            Cells[p.X,p.Y]:=savecell;
            grid1.update;
            If progressbox.checked then sleep(1000);
            result:=placeNextStar(starnbr+1, nextgrid);
            if result then exit; {good placement, return}
          end ;
          {placement invalid, retract it}
          nextgrid[p.X,p.Y]:=starNbr;  {Remove the guess}
          grid1.cells[p.X,p.Y]:=savecell; {fix grid display}
          grid1.update;
          If progressbox.checked then sleep(1000);
          memo1.lines.add(format('Figure %d loc(%d,%d)incorrect,retracted',
                                        [starnbr, p.x,p.y]));

        end;
      end;
    end;

begin
  redrawgrid;
  if not PlaceNextStar(0, grid)
  then showmessage('No solution found');
end;

end.
